@@ -1,26 +1,6 @@
-/* Spa
- *
- * Copyright © 2018 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* Spa */
+/* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #include <unistd.h>
 #include <errno.h>
@@ -89,7 +69,7 @@ struct impl {
 	uint8_t *buffer_data;
 	uint8_t buffer_mem[DATAS_SIZE + MAX_ALIGN];
 
-	unsigned int flushing:1;
+	uint32_t flush_count;
 	unsigned int polling:1;
 };
 
@@ -166,44 +146,61 @@ static int loop_remove_source(void *object, struct spa_source *source)
 
 static void flush_items(struct impl *impl)
 {
-	uint32_t index;
+	uint32_t index, flush_count;
+	int32_t avail;
 	int res;
 
-	impl->flushing = true;
-	while (spa_ringbuffer_get_read_index(&impl->buffer, &index) > 0) {
+	flush_count = ++impl->flush_count;
+	avail = spa_ringbuffer_get_read_index(&impl->buffer, &index);
+	while (avail > 0) {
 		struct invoke_item *item;
 		bool block;
+		spa_invoke_func_t func;
 
 		item = SPA_PTROFF(impl->buffer_data, index & (DATAS_SIZE - 1), struct invoke_item);
 		block = item->block;
+		func = item->func;
 
 		spa_log_trace_fp(impl->log, "%p: flush item %p", impl, item);
-		item->res = item->func ? item->func(&impl->loop,
-				true, item->seq, item->data, item->size,
-			   item->user_data) : 0;
+		/* first we remove the function from the item so that recursive
+		 * calls don't call the callback again. We can't update the
+		 * read index before we call the function because then the item
+		 * might get overwritten. */
+		item->func = NULL;
+		if (func)
+			item->res = func(&impl->loop, true, item->seq, item->data,
+				item->size, item->user_data);
 
-		spa_ringbuffer_read_update(&impl->buffer, index + item->item_size);
+		/* if this function did a recursive invoke, it now flushed the
+		 * ringbuffer and we can exit */
+		if (flush_count != impl->flush_count)
+			break;
+
+		index += item->item_size;
+		avail -= item->item_size;
+		spa_ringbuffer_read_update(&impl->buffer, index);
 
 		if (block) {
 			if ((res = spa_system_eventfd_write(impl->system, impl->ack_fd, 1)) < 0)
-				spa_log_warn(impl->log, "%p: failed to write event fd: %s",
-						impl, spa_strerror(res));
+				spa_log_warn(impl->log, "%p: failed to write event fd:%d: %s",
+						impl, impl->ack_fd, spa_strerror(res));
 		}
 	}
-	impl->flushing = false;
 }
 
 static int
 loop_invoke_inthread(struct impl *impl,
-	    spa_invoke_func_t func,
-	    uint32_t seq,
-	    const void *data,
-	    size_t size,
-	    bool block,
-	    void *user_data)
+		spa_invoke_func_t func,
+		uint32_t seq,
+		const void *data,
+		size_t size,
+		bool block,
+		void *user_data)
 {
-	if (!impl->flushing)
-		flush_items(impl);
+	/* we should probably have a second ringbuffer for the in-thread pending
+	 * callbacks. A recursive callback when flushing will insert itself
+	 * before this one. */
+	flush_items(impl);
 	return func ? func(&impl->loop, true, seq, data, size, user_data) : 0;
 }
 
@@ -222,6 +219,9 @@ loop_invoke(void *object,
 	int32_t filled;
 	uint32_t avail, idx, offset, l0;
 
+	/* the ringbuffer can only be written to from one thread, if we are
+	 * in the same thread as the loop, don't write into the ringbuffer
+	 * but try to emit the calback right away after flushing what we have */
 	if (impl->thread == 0 || pthread_equal(impl->thread, pthread_self()))
 		return loop_invoke_inthread(impl, func, seq, data, size, block, user_data);
 
@@ -247,6 +247,7 @@ loop_invoke(void *object,
 	item->size = size;
 	item->block = block;
 	item->user_data = user_data;
+	item->res = 0;
 	item->item_size = SPA_ROUND_UP_N(sizeof(struct invoke_item) + size, ITEM_ALIGN);
 
 	spa_log_trace_fp(impl->log, "%p: add item %p filled:%d", impl, item, filled);
@@ -283,8 +284,8 @@ loop_invoke(void *object,
 		spa_loop_control_hook_before(&impl->hooks_list);
 
 		if ((res = spa_system_eventfd_read(impl->system, impl->ack_fd, &count)) < 0)
-			spa_log_warn(impl->log, "%p: failed to read event fd: %s",
-					impl, spa_strerror(res));
+			spa_log_warn(impl->log, "%p: failed to read event fd:%d: %s",
+					impl, impl->ack_fd, spa_strerror(res));
 
 		spa_loop_control_hook_after(&impl->hooks_list);
 
@@ -318,6 +319,8 @@ loop_add_hook(void *object,
 	      void *data)
 {
 	struct impl *impl = object;
+	spa_return_if_fail(SPA_CALLBACK_CHECK(hooks, before, 0));
+	spa_return_if_fail(SPA_CALLBACK_CHECK(hooks, after, 0));
 	spa_hook_list_append(&impl->hooks_list, hook, hooks, data);
 }
 
@@ -332,10 +335,10 @@ static void loop_enter(void *object)
 		impl->enter_count = 1;
 	} else {
 		spa_return_if_fail(impl->enter_count > 0);
-		spa_return_if_fail(impl->thread == thread_id);
+		spa_return_if_fail(pthread_equal(impl->thread, thread_id));
 		impl->enter_count++;
 	}
-	spa_log_trace(impl->log, "%p: enter %lu", impl, impl->thread);
+	spa_log_trace_fp(impl->log, "%p: enter %p", impl, (void *) impl->thread);
 }
 
 static void loop_leave(void *object)
@@ -344,16 +347,24 @@ static void loop_leave(void *object)
 	pthread_t thread_id = pthread_self();
 
 	spa_return_if_fail(impl->enter_count > 0);
-	spa_return_if_fail(impl->thread == thread_id);
+	spa_return_if_fail(pthread_equal(impl->thread, thread_id));
 
-	spa_log_trace(impl->log, "%p: leave %lu", impl, impl->thread);
+	spa_log_trace_fp(impl->log, "%p: leave %p", impl, (void *) impl->thread);
 
 	if (--impl->enter_count == 0) {
 		impl->thread = 0;
+		flush_items(impl);
 		impl->polling = false;
 	}
 }
 
+static int loop_check(void *object)
+{
+	struct impl *impl = object;
+	pthread_t thread_id = pthread_self();
+	return (impl->thread == 0 || pthread_equal(impl->thread, thread_id)) ? 1 : 0;
+}
+
 static inline void free_source(struct source_impl *s)
 {
 	detach_source(&s->source);
@@ -388,7 +399,7 @@ static void cancellation_handler(void *closure)
 	}
 }
 
-static int loop_iterate(void *object, int timeout)
+static int loop_iterate_cancel(void *object, int timeout)
 {
 	struct impl *impl = object;
 	struct spa_poll_event ep[MAX_EP], *e;
@@ -435,6 +446,52 @@ static int loop_iterate(void *object, int timeout)
 	return nfds;
 }
 
+static int loop_iterate(void *object, int timeout)
+{
+	struct impl *impl = object;
+	struct spa_poll_event ep[MAX_EP], *e;
+	int i, nfds;
+
+	impl->polling = true;
+	spa_loop_control_hook_before(&impl->hooks_list);
+
+	nfds = spa_system_pollfd_wait(impl->system, impl->poll_fd, ep, SPA_N_ELEMENTS(ep), timeout);
+
+	spa_loop_control_hook_after(&impl->hooks_list);
+	impl->polling = false;
+
+	/* first we set all the rmasks, then call the callbacks. The reason is that
+	 * some callback might also want to look at other sources it manages and
+	 * can then reset the rmask to suppress the callback */
+	for (i = 0; i < nfds; i++) {
+		struct spa_source *s = ep[i].data;
+
+		s->rmask = ep[i].events;
+		/* already active in another iteration of the loop,
+		 * remove it from that iteration */
+		if (SPA_UNLIKELY(e = s->priv))
+			e->data = NULL;
+		s->priv = &ep[i];
+	}
+
+	if (SPA_UNLIKELY(!spa_list_is_empty(&impl->destroy_list)))
+		process_destroy(impl);
+
+	for (i = 0; i < nfds; i++) {
+		struct spa_source *s = ep[i].data;
+		if (SPA_LIKELY(s && s->rmask))
+			s->func(s);
+	}
+	for (i = 0; i < nfds; i++) {
+		struct spa_source *s = ep[i].data;
+		if (SPA_LIKELY(s)) {
+			s->rmask = 0;
+			s->priv = NULL;
+		}
+	}
+	return nfds;
+}
+
 static void source_io_func(struct spa_source *source)
 {
 	struct source_impl *s = SPA_CONTAINER_OF(source, struct source_impl, source);
@@ -525,12 +582,12 @@ static int loop_enable_idle(void *object, struct spa_source *source, bool enable
 
 	if (enabled && !s->enabled) {
 		if ((res = spa_system_eventfd_write(s->impl->system, source->fd, 1)) < 0)
-			spa_log_warn(s->impl->log, "%p: failed to write idle fd %d: %s",
+			spa_log_warn(s->impl->log, "%p: failed to write idle fd:%d: %s",
 					source, source->fd, spa_strerror(res));
 	} else if (!enabled && s->enabled) {
 		uint64_t count;
 		if ((res = spa_system_eventfd_read(s->impl->system, source->fd, &count)) < 0)
-			spa_log_warn(s->impl->log, "%p: failed to read idle fd %d: %s",
+			spa_log_warn(s->impl->log, "%p: failed to read idle fd:%d: %s",
 					source, source->fd, spa_strerror(res));
 	}
 	s->enabled = enabled;
@@ -584,10 +641,12 @@ static void source_event_func(struct spa_source *source)
 	uint64_t count = 0;
 	int res;
 
-	if ((res = spa_system_eventfd_read(s->impl->system, source->fd, &count)) < 0)
-		spa_log_warn(s->impl->log, "%p: failed to read event fd %d: %s",
-				source, source->fd, spa_strerror(res));
-
+	if ((res = spa_system_eventfd_read(s->impl->system, source->fd, &count)) < 0) {
+		if (res != -EAGAIN)
+			spa_log_warn(s->impl->log, "%p: failed to read event fd:%d: %s",
+					source, source->fd, spa_strerror(res));
+		return;
+	}
 	s->func.event(source->data, count);
 }
 
@@ -638,7 +697,7 @@ static int loop_signal_event(void *object, struct spa_source *source)
 	spa_assert(source->func == source_event_func);
 
 	if (SPA_UNLIKELY((res = spa_system_eventfd_write(s->impl->system, source->fd, 1)) < 0))
-		spa_log_warn(s->impl->log, "%p: failed to write event fd %d: %s",
+		spa_log_warn(s->impl->log, "%p: failed to write event fd:%d: %s",
 				source, source->fd, spa_strerror(res));
 	return res;
 }
@@ -650,10 +709,12 @@ static void source_timer_func(struct spa_source *source)
 	int res;
 
 	if (SPA_UNLIKELY((res = spa_system_timerfd_read(s->impl->system,
-				source->fd, &expirations)) < 0))
-		spa_log_warn(s->impl->log, "%p: failed to read timer fd %d: %s",
-				source, source->fd, spa_strerror(res));
-
+				source->fd, &expirations)) < 0)) {
+		if (res != -EAGAIN)
+			spa_log_warn(s->impl->log, "%p: failed to read timer fd:%d: %s",
+					source, source->fd, spa_strerror(res));
+		return;
+	}
 	s->func.timer(source->data, expirations);
 }
 
@@ -730,10 +791,12 @@ static void source_signal_func(struct spa_source *source)
 	struct source_impl *s = SPA_CONTAINER_OF(source, struct source_impl, source);
 	int res, signal_number = 0;
 
-	if ((res = spa_system_signalfd_read(s->impl->system, source->fd, &signal_number)) < 0)
-		spa_log_warn(s->impl->log, "%p: failed to read signal fd %d: %s",
-				source, source->fd, spa_strerror(res));
-
+	if ((res = spa_system_signalfd_read(s->impl->system, source->fd, &signal_number)) < 0) {
+		if (res != -EAGAIN)
+			spa_log_warn(s->impl->log, "%p: failed to read signal fd:%d: %s",
+					source, source->fd, spa_strerror(res));
+		return;
+	}
 	s->func.signal(source->data, signal_number);
 }
 
@@ -811,6 +874,16 @@ static const struct spa_loop_methods impl_loop = {
 	.invoke = loop_invoke,
 };
 
+static const struct spa_loop_control_methods impl_loop_control_cancel = {
+	SPA_VERSION_LOOP_CONTROL_METHODS,
+	.get_fd = loop_get_fd,
+	.add_hook = loop_add_hook,
+	.enter = loop_enter,
+	.leave = loop_leave,
+	.iterate = loop_iterate_cancel,
+	.check = loop_check,
+};
+
 static const struct spa_loop_control_methods impl_loop_control = {
 	SPA_VERSION_LOOP_CONTROL_METHODS,
 	.get_fd = loop_get_fd,
@@ -818,6 +891,7 @@ static const struct spa_loop_control_methods impl_loop_control = {
 	.enter = loop_enter,
 	.leave = loop_leave,
 	.iterate = loop_iterate,
+	.check = loop_check,
 };
 
 static const struct spa_loop_utils_methods impl_loop_utils = {
@@ -864,11 +938,9 @@ static int impl_clear(struct spa_handle *handle)
 
 	impl = (struct impl *) handle;
 
-	if (impl->enter_count != 0)
-		spa_log_warn(impl->log, "%p: loop is entered %d times",
-				impl, impl->enter_count);
-
-	spa_assert(!impl->polling);
+	if (impl->enter_count != 0 || impl->polling)
+		spa_log_warn(impl->log, "%p: loop is entered %d times polling:%d",
+				impl, impl->enter_count, impl->polling);
 
 	spa_list_consume(source, &impl->source_list, link)
 		loop_destroy_source(impl, &source->source);
@@ -894,6 +966,7 @@ impl_init(const struct spa_handle_factory *factory,
 	  uint32_t n_support)
 {
 	struct impl *impl;
+	const char *str;
 	int res;
 
 	spa_return_val_if_fail(factory != NULL, -EINVAL);
@@ -916,6 +989,12 @@ impl_init(const struct spa_handle_factory *factory,
 			SPA_VERSION_LOOP_UTILS,
 			&impl_loop_utils, impl);
 
+	if (info) {
+		if ((str = spa_dict_lookup(info, "loop.cancel")) != NULL &&
+		    spa_atob(str))
+			impl->control.iface.cb.funcs = &impl_loop_control_cancel;
+	}
+
 	impl->log = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_Log);
 	spa_log_topic_init(impl->log, &log_topic);
 	impl->system = spa_support_find(support, n_support, SPA_TYPE_INTERFACE_System);
